#####
# .mwparms (MegaWizard Paramters) file manipulation routines
#####
namespace eval ::altera::mwparms {

	# parse line of mwparms file looking for parameter:
	#  parm=val with some or no whitespace surrounding '='
	# return: on success: list <key> <val> 
	#			 on failure: list inLine
	proc parseLine { inLine } {
		if {[regexp {(^\S+)\s*=\s*(.*$)} $inLine whole key val]} {
			return [list $key $val]
		}
		return [list $inLine]
	}
	
	# write line to mwparms file fd as:
	# <key>=<val>
	proc writeLine { fd lineList } {
		set key [string trim [lindex $lineList 0]]
		set val [string trim [lindex $lineList 1]]
		puts $fd "$key=$val"
	}
	
	# Take in a key-value list for replacing values in inFileName writing result
	#  to outFileName(currently diabled) and returning in outputParamArrIn
	proc setParams { inFileName outFileName paramArrIn outputParamArrIn} {
		upvar $paramArrIn paramArr
		upvar $outputParamArrIn outputParamArr
		
		set fin [open $inFileName r]
		#set fout [open $outFileName w]
		while {[gets $fin line] >= 0} {
			# pLin = list: <key> <val>
			set pLin [parseLine $line] 
			if {[llength $pLin] > 1} {
				# check to see if key is in set to change value
				set key [lindex $pLin 0]
				if {[info exists paramArr($key)]} {
					set keyVal $paramArr($key)
					lset pLin 1 $keyVal
					#writeLine $fout $pLin
				} else {
					# not in list, output line as-is
					#puts $fout "$line"
				}
				# store final mwparms paramters for caller
				set outputParamArr([lindex $pLin 0]) [lindex $pLin 1]
			} else {
				# nothing to check, output line as-is
				#puts $fout "$line"
			}
			
		}
		close $fin
		#close $fout
	}
	
}

return 0