# NOTE: Output to stdout during load OR ANY Tcl error may cause errors in impulse_sm (or Stage master Explorer)
#			and it will spin and never exit (or sometimes bring up the Stage Master Explorer window)
namespace eval ::altera::floating_point {
	namespace export hdl*
	namespace export gen*
	namespace export convert*

	array set GeneratedOpArr {}

	namespace eval constants {
		#####
		# namespace constants
		variable PSPCOMMON_PATH [file join $::env(IMPULSEC_HOME) Architectures/PSPcommon/Tcl]
		variable MEGAWIZARD_COM_PATH [file join $::env(IMPULSEC_HOME) Architectures/Altera/AutoIP/Common/MegaWizard]
		variable MEGAWIZARD_FP_PATH [file join $::env(IMPULSEC_HOME) Architectures/Altera/AutoIP/FloatingPoint/MegaWizard]
		variable FP_OP_LIST_FILENAME "fp_op_list.txt"

		# Map XML paramter values to MegaWizard generic values
		# <xml param name> { name <xml param name> [<xml val> <MegaWizard val>...]}
		variable XML_TO_MWGENERIC_VALMAP
		array set XML_TO_MWGENERIC_VALMAP {
			latency {
				name PIPELINE
			}
			fxpt_iwidth {
				name {WIDTH_INT WIDTH_DATA}
			}
			fxpt_uwidth {
				name {WIDTH_INT WIDTH_DATA}
				expression "[calcWidth a paramArr]"
			}
			fxpt_owidth {
				name {WIDTH_INT WIDTH_RESULT}
			}
		}
		# Map XML paramter values to MegaWizard parameter values for generating core and simulation files
		# <xml param name> { name <MegaWizard param name> [<xml val> <generic val>...]}
		variable XML_TO_MWPARM_VALMAP
		array set XML_TO_MWPARM_VALMAP {
			latency {
				name PIPELINE
			}
			fxpt_iwidth {
				name {WIDTH_INT WIDTH_DATA}
			}
			fxpt_uwidth {
				name {WIDTH_INT WIDTH_DATA}
				expression "[calcWidth a paramArr]"
			}
			fxpt_owidth {
				name {WIDTH_INT WIDTH_RESULT}
			}
		}

		variable OPNAME_TO_MWPARMS_NAME_MAP
		array set OPNAME_TO_MWPARMS_NAME_MAP {
			fadd altfp_add
			faddd altfp_addd
			fsub altfp_sub
			fsubd altfp_subd
			fmul altfp_mult
			fmuld altfp_multd
			fdiv altfp_div
			fdivd altfp_divd
			feq altfp_cmp_eq
			feqd altfp_cmp_eqd
			fneq altfp_cmp_neq
			fneqd altfp_cmp_neqd
			flt altfp_cmp_lt
			fltd altfp_cmp_ltd
			flteq altfp_cmp_lte
			flteqd altfp_cmp_lted
			itof altfp_itof
			itod altfp_itod
			utof altfp_utof
			utod altfp_utod
			ftoi altfp_ftoi
			dtoi altfp_dtoi
			fsqrt altfp_sqrt
			fsqrtd altfp_sqrtd
			flog altfp_log
			flogd altfp_logd
			fexp altfp_exp
			fexpd altfp_expd
			fcos altfp_cos
			fsin altfp_sin
			fatan altfp_atan
		}

		# Note: "float" can mean either "float 32-bit" or "double 64-bit" for IO
		# NOTE: expressions may use local variables that must be present in resolving proc
		#		example: $paramArr(width) uses array 'paramArr' local to proc calling 'subst'
		# HDL functions for 2-float-in-1-float-out modules (e.g. fmul)
		namespace eval twoFloatInOneFloatOut {
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
			}
			# <port name> <width expr>
			variable DEF_PORTS {
    			{"a" $paramArr(width)}
    			{"b" $paramArr(width)}
    			{"pipeEn" 0}
    			{"result" $paramArr(width)}
			}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				{clock 0 0 std_logic}
				{dataa 0 $paramArr(width) std_logic_vector}
				{datab 0 $paramArr(width) std_logic_vector}
				{clk_en 0 0 std_logic}
				{result 1 $paramArr(width) std_logic_vector}
			}
			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				clock clk
				dataa a
				datab b
				clk_en pipeEn
				result result
			}
		}
		# HDL functions for 1-float-in-1-float-out modules (e.g. sqrt())
		namespace eval oneFloatInOneFloatOut {
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
			}
			# <port name> <width expr>
			variable DEF_PORTS {
				{"a" $paramArr(width)}
				{"pipeEn" 0}
				{"result" $paramArr(width)}
			}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				{clock 0 0 std_logic}
				{data 0 $paramArr(width) std_logic_vector}
				{clk_en 0 0 std_logic}
				{result 1 $paramArr(width) std_logic_vector}
			}
			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				clock clk
				data a
				clk_en pipeEn
				result result
			}
		}
		# HDL functions for 2-float-in-1-bit-out modules (e.g. comparisons)
		namespace eval twoFloatInOneBitOut {
			namespace export *
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
				{"core_result" 0}
			}
			# <port name> <width expr>
			variable DEF_PORTS {
				{"a" $paramArr(width)}
				{"b" $paramArr(width)}
				{"pipeEn" 0}
				{"result" 1}
			}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				{clock 0 0 std_logic}
				{dataa 0 $paramArr(width) std_logic_vector}
				{datab 0 $paramArr(width) std_logic_vector}
				{clk_en 0 0 std_logic}
				{$opNameToResultPortMap($opName) 1 0 std_logic}
			}

			variable OPNAME_TO_RESULT_PORT_MAP
			array set OPNAME_TO_RESULT_PORT_MAP {
				feq aeb
				feqd aeb
				flt alb
				fltd alb
				flteq aleb
				flteqd aleb
				fneq aneb
				fneqd aneb
			}
			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				clock clk
				dataa a
				datab b
				clk_en pipeEn
				$opNameToResultPortMap($opName) core_result
			}
			proc genIpForOpIOType {fd opName opIOType lang compName paramArrIn} {
				hdl::assign "result" [hdl::val "(0=>core_result)"]
			}
		}
		# HDL functions for 1-float-in-1-int-out modules (e.g. float-to-int)
		namespace eval oneFloatInOneIntOut {
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
			}
			# <port name> <width expr>
			variable DEF_PORTS {
				{"a" $paramArr(width)}
				{"pipeEn" 0}
				{"result" $paramArr(fxpt_owidth)}
			}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				{clock 0 0 std_logic}
				{dataa 0 $paramArr(width) std_logic_vector}
				{clk_en 0 0 std_logic}
				{result 1 $paramArr(fxpt_owidth) std_logic_vector}
			}
			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				clock clk
				dataa a
				clk_en pipeEn
				result result
			}
		}
		# HDL functions for 1-int-in-1-float-out modules (e.g. int-to-float)
		namespace eval oneIntInOneFloatOut {
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
			}
			# <port name> <width expr>
			variable DEF_PORTS {
				  {"a" $paramArr(fxpt_iwidth)}
				  {"pipeEn" 0}
				  {"result" $paramArr(width)}
			  }
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				  {clock 0 0 std_logic}
				  {dataa 0 $paramArr(fxpt_iwidth) std_logic_vector}
				  {clk_en 0 0 std_logic}
				  {result 1 $paramArr(width) std_logic_vector}
			  }
			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				  clock clk
				  dataa a
				  clk_en pipeEn
				  result result
  			  }
		}
		# HDL functions for 1-unsignedInt-in-1-float-out modules (e.g. int-to-float)
		namespace eval oneUIntInOneFloatOut {
			namespace export *
			# <name> <width> [init]
			variable DEF_GENERICS {
			}
			# <name> <width> [type]
			variable DEF_SIGNALS {
				{"core_result" $paramArr(width)}
			}
			# <port name> <width expr>
			variable DEF_PORTS {
				{"a" $paramArr(fxpt_uwidth)}
				{"pipeEn" 0}
				{"result" $paramArr(width)}
			}
			# <port name> <0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]
			variable DEF_COMP_PORTS {
				{clock 0 0 std_logic}
				{dataa 0 [calcWidth a paramArr] std_logic_vector}
				{clk_en 0 0 std_logic}
				{result 1 $paramArr(width) std_logic_vector}
			}

			proc calcWidth {sigName paramArrIn} {
				upvar $paramArrIn paramArr
				switch $sigName {
					{a} {
						if {$paramArr(fxpt_uwidth) < 64} {
							return [expr $paramArr(fxpt_uwidth) + 1]
						} else {
							return 64
						}
					}
					default {
						::altera::floating_point::fatalError "calcWidth:Don't know how to calculate width for \[$sigName\]"
					}
				}
			}

			# <comp port> <signal name>
			variable COMP_INST_PORT_MAP {
				clock clk
				dataa [calcCompInstPortWidth a paramArr]
				clk_en pipeEn
				result core_result
			}

			proc calcCompInstPortWidth {sigName paramArrIn} {
				upvar $paramArrIn paramArr
				switch $sigName {
					{a} {
						if {$paramArr(fxpt_uwidth) < 64} {
							#return {"'0' & a"}
							return "alt64unsigned"
						} else {
							return "a64unsigned"
						}
					}
					default {
						::altera::floating_point::fatalError "calcCompInstPortWidth:Don't know how to calculate width for \[$sigName\]"
					}
				}
			}

			# routine specific to this opIOType for generating additional HDL
			# utof/utod functions
			proc genIpForOpIOType {fd opName opIOType lang compName paramArrIn} {
				upvar $paramArrIn paramArr

				# utod must be scaled if input is 64 bits because utod core only accepts signed
				#  integers.  If input width is < 64 then do nothing, at 64 bits scale input/output
				#  by 2 only if the MSb is set.  >64 bits is unsupported
				set awidth $paramArr(fxpt_uwidth)
				if {$awidth > 64} {
					::altera::floating_point::fatalError "$opName of >64 bits is not supported, use correct type or cast to <=64 bit"
				} elseif {$awidth < 64} {
					# <64 bit input, just assign
					hdl::defineSig "alt64unsigned" [calcWidth a paramArr]
					hdl::assign "alt64unsigned" [hdl::cat [hdl::bit 0] [hdl::val "a"]]
					hdl::assign "result" [hdl::val "core_result"]
				} else {
					# 64-bit input, need to instantiate logic that does scale by 2 based on MSb
					set latency $paramArr(latency)
					# additional signals
					hdl::defineSig "a64unsigned" [calcWidth a paramArr]
					hdl::defineSig "msb_dly" 1
					hdl::defineSig "msb" 1
					hdl::defineSig "newexp" 11

					# assignments based on MSb
					hdl::assign "msb" [hdl::subvec "a" [expr $awidth - 1] [expr $awidth - 1]]

					hdl::assign "a64unsigned" [hdl::cat [hdl::bit 0] [hdl::subvec "a" [expr $awidth - 1] 1]] [hdl::eq [hdl::elm "a" [expr $awidth - 1]] [hdl::bit 1]]
					hdl::assign "a64unsigned" [hdl::val "a"]

					hdl::assign "newexp" [hdl::add [hdl::subvec "core_result" 62 52] [hdl::cat [hdl::const "0000000000"] [hdl::val msb_dly]]]

					hdl::assign "result" [hdl::cat [hdl::bit 0] [hdl::cat [hdl::val "newexp"] [hdl::subvec "core_result" 51 0]]]

					#hdl::assign "result" [hdl::cat "0" [hdl::cat "std_ulogic_vector(expxtwo)" "std_ulogic_vector(result_slv(51 downto 0))"]]

					# delay element component
					#hdl::defineLibrary "impulse"
					hdl::defineComponent latency_delay [list [list LATENCY integer 1]\
						[list WIDTH integer 1]]\
						[list [list clk 0 0]\
						[list a 0 1]\
						[list pipeEn 0 0]\
						[list a_dly 1 1]]
					# delay element instance
					set genericActuals "[list LATENCY [hdl::val $latency] WIDTH [hdl::val 1]]"
					#[list LATENCY $latency WIDTH 1]
					puts "genericActuals=$genericActuals"
					hdl::instance latency_delay_inst latency_delay $genericActuals \
						[list clk clk a msb pipeEn pipeEn a_dly msb_dly]
				}
			}
		}

		# import constants into caller's space as local variables with same names
		proc import {args} {
			set nmToImport [linsert $args 0 ""]
			foreach {subConstNS} $nmToImport {
				if {$subConstNS != ""} {
					set subConstNS "::${subConstNS}"
				}
				#puts "importing sub-constant namespace \[$subConstNS\]"
				foreach {const} [info vars "[namespace current]${subConstNS}::*"] {
					#puts "importing $const"
					uplevel 1 [list variable "$const"]
				}
			}
		}

	}

	#load dependency Tcl
	proc init {} {
		source [file join $::altera::floating_point::constants::MEGAWIZARD_COM_PATH mwparms.tcl]
		source [file join $::altera::floating_point::constants::MEGAWIZARD_COM_PATH mwinst.tcl]

		if { [file exists $::altera::floating_point::constants::FP_OP_LIST_FILENAME] } {
			file delete $::altera::floating_point::constants::FP_OP_LIST_FILENAME
		}
	}

	proc hdlDefineGenerics {paramArrIn defGenericsList} {
		upvar $paramArrIn paramArr
		foreach {genericDef} $defGenericsList {
			puts "Defining generic $genericDef"
			eval hdl::defineGeneric [join $genericDef]
		}
	}

	proc hdlDefineSignals {paramArrIn defSignalsList} {
		upvar $paramArrIn paramArr
		foreach {signalDef} $defSignalsList {
			eval hdl::defineSig [join $signalDef]
		}
	}

	proc hdlDefinePorts {paramArrIn defPortList} {
		upvar $paramArrIn paramArr
		foreach {portDef} $defPortList {
			eval hdl::definePort [join $portDef]
		}
	}

	proc hdlDefineComponent {compName paramArrIn defCompPortsList generics opName opNameToResultPortMapIn} {
		upvar $paramArrIn paramArr
		upvar $opNameToResultPortMapIn opNameToResultPortMap

		set ports {}
		foreach {portDef} $defCompPortsList {
			lappend ports [subst $portDef]
		}
		hdl::defineComponent $compName $generics $ports
	}

	proc hdlInstanceComponent {compInstName paramArrIn portMapList opName opNameToResultPortMapIn} {
		upvar $paramArrIn paramArr
		upvar $opNameToResultPortMapIn opNameToResultPortMap

		set substList [subst $portMapList]
		#puts "substList=$substList"
		hdl::instance "${compInstName}_inst" $compInstName {} $substList
	}

	proc fatalError {errStr} {
		puts ""
		puts "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
		puts "!! FATAL_ERROR:"
		puts "!!   $errStr"
		puts "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
		puts "Exiting on fatal error..."
		exit -1;
	}

	# apply translation array on parameter list
	proc applyTranslationArray { paramList transArrIn} {
		upvar $transArrIn transArr
		array set paramArr $paramList
		array set outParams {}
		foreach {paramKey paramVal} $paramList {
			#puts "mapping $paramKey $paramVal"
			if {[info exists transArr($paramKey)]} {
				array set xmlVal2TransArr $transArr($paramKey)
				# names can be single or list
				set transKeyList $xmlVal2TransArr(name)
				foreach transKey $transKeyList {
					if {[info exists xmlVal2TransArr($paramVal)]} {
						# translate by value if present
						set transVal $xmlVal2TransArr($paramVal)
					} elseif {[info exists xmlVal2TransArr(expression)]} {
						# translate by "expression" if present
						set transVal [subst $xmlVal2TransArr(expression)]
					} else {
						# no translation, use incoming value
						set transVal $paramVal
					}
					set outParams($transKey) $transVal
				}
				# unset for clean next lookup
				unset xmlVal2TransArr
			}
		}
		return [array get outParams]
	}

	# lookup, and translate if necessary, XML parameters into generics
	proc convertXmlToMwGenericParameters {xmlParamList} {
		constants::import
		return [applyTranslationArray $xmlParamList XML_TO_MWGENERIC_VALMAP]
	}

	# lookup, and translate if necessary, XML parameters into MegaWizard parameters
	proc convertXmlToMwparmsParameters {xmlParamList} {
		constants::import
		return [applyTranslationArray $xmlParamList XML_TO_MWPARM_VALMAP]
	}

	# convert the DEF_COMP_PORTS into a list ready to become an array of <port name>=<list>
	#  also resolves values in <list>
	proc convertCompPortListToArrayList {paramArrIn defCompPortsList} {
		upvar $paramArrIn paramArr
		set arrList [list]
		foreach {portDef} $defCompPortsList {
			set resolved [subst $portDef]
			lappend arrList [lindex $resolved 0] [lrange $resolved 1 end]
		}
		#		puts "arrList=$arrList"
		return $arrList
	}

	proc printComment {fd lang comment} {
		if {$lang == "vhdl"} {
			set commTxt "-- "
		} else {
			set commTxt "// "
		}
		puts $fd "${commTxt}${comment}"
	}

	proc writeLine {fd tabLvl str} {
		puts $fd "[string repeat {   } $tabLvl]$str"
	}

	proc genHdlComponent {fd lang compName instName portsList genericsList tabLvl} {

		writeLine $fd $tabLvl "component ${instName} is"
		incr tabLvl

		if { [llength $genericsList] > 0 } {
			writeLine $fd $tabLvl "generic ("
			incr tabLvl
			set numGenerics [llength $genericsList]
			set genericNum 0
			foreach generic $genericsList {
				set lineStr "[lindex $generic 0] : "
				if { [string is integer [lindex $generic 1] ] } {
					append lineStr "Integer"
				} else {
					append lineStr "String"
				}
				incr genericNum
				if { $genericNum != $numGenerics } {
					append lineStr ";"
				}
				writeLine $fd $tabLvl "$lineStr"
			}
			incr tabLvl -1
			writeLine $fd $tabLvl ");"
		}

		writeLine $fd $tabLvl "port ("
		incr tabLvl
		set numPorts [llength $portsList]
		set portNum 0
		foreach port $portsList {
			set lineStr "[lindex $port 0] : "
			if { [lindex $port 1] == 0 } {
				append lineStr "in "
			} else {
				append lineStr "out "
			}
			if { [lindex $port 2] == 0 || [lindex $port 2] == 1} {
				#append lineStr "[lindex $port 3]"
				append lineStr "std_logic"
			} else {
				append lineStr "[lindex $port 3]([lindex $port 2]-1 downto 0)"
			}
			incr portNum
			if { $portNum != $numPorts } {
				append lineStr ";"
			}
			writeLine $fd $tabLvl $lineStr
		}
		incr tabLvl -1
		writeLine $fd $tabLvl ");"
		incr tabLvl -1
		writeLine $fd $tabLvl "end component;"

	}

	proc genHdlInstance {fd lang compName instName portsList genericsList tabLvl} {

		writeLine $fd $tabLvl "${instName}_inst : $instName"
		incr tabLvl

		if { [llength $genericsList] > 0 } {
			writeLine $fd $tabLvl "generic map ("
			incr tabLvl
			set numGenerics [llength $genericsList]
			set genericNum 0
			foreach generic $genericsList {
				set lineStr "[lindex $generic 0] => [lindex $generic 1]"
				incr genericNum
				if { $genericNum != $numGenerics } {
					append lineStr ","
				}
				writeLine $fd $tabLvl "$lineStr"
			}
			incr tabLvl -1
			writeLine $fd $tabLvl ")"
		}

		writeLine $fd $tabLvl "port map ("
		incr tabLvl
		set numPorts [llength $portsList]
		set portNum 0
		foreach port $portsList {
			set lineStr "[lindex $port 0] => [lindex $port 0]"
			# sort of kludgy to handle signal-to-vector(0 downto 0) conversion
			if { [lindex $port 2] == 1 } {
				append lineStr "(0)"
			}
			incr portNum
			if { $portNum != $numPorts } {
				append lineStr ","
			}
			writeLine $fd $tabLvl "$lineStr"
		}
		incr tabLvl -1
		writeLine $fd $tabLvl ");"
		incr tabLvl -1

	}

	proc genHdlWrapperForOpComp {fd lang compName instName paramArrIn portsList genericsList opName opNameToResultPortMapIn mwparmsOutputParamsArrIn} {
		upvar $paramArrIn paramArr
		upvar $opNameToResultPortMapIn opNameToResultPortMap
		upvar $mwparmsOutputParamsArrIn mwparmsOutputParamsArr

		# As of ~v14.0, Quartus has issues generating altfp_* from VHDL due to LPM library using AHDL/TDF files.
		#  Now MegaWizard must be used to generated both synthesis and simulation files

		printComment $fd $lang "$compName referenced below needs to be generated via MegaWizard using the command:"

		set tabLvl 0
		set lineStr "-- qmegawiz -silent module=\"$instName\" "
		foreach {key val} [array get mwparmsOutputParamsArr] {
			append lineStr "$key=$val "
		}
		append lineStr "-p:work_dir ${compName}.vhd"
		writeLine $fd $tabLvl $lineStr

		# EARLY RETURN
		if {1} return

		set tabLvl 0

		set portListSubst [subst $portsList]

		printComment $fd $lang "$compName is a wrapper for specialized instance and instantiated below"

		puts $fd "-- synthesis will use the following component, simulation will have a model created via MegaWizard"

		#libraries
		puts $fd "library ieee;
use ieee.std_logic_1164.all;
LIBRARY altera_mf;
USE altera_mf.altera_mf_components.all;
use work.all;
"
		# entity
		writeLine $fd $tabLvl "entity $compName is"
		incr tabLvl
		# generics

		# ports
		writeLine $fd $tabLvl "port ("
		incr tabLvl
		set numPorts [llength $portListSubst]
		set portNum 0
		foreach port $portListSubst {
			set lineStr "[lindex $port 0] : "
			if { [lindex $port 1] == 0 } {
				append lineStr "in "
			} else {
				append lineStr "out "
			}
			if { [lindex $port 2] == 0 } {
				append lineStr "[lindex $port 3]"
			} else {
				append lineStr "[lindex $port 3]([lindex $port 2]-1 downto 0)"
			}
			incr portNum
			if { $portNum != $numPorts } {
				append lineStr ";"
			}
			writeLine $fd $tabLvl $lineStr
		}
		incr tabLvl -1
		writeLine $fd $tabLvl ");"
		incr tabLvl -1
		writeLine $fd $tabLvl "end;"
		writeLine $fd $tabLvl ""

		# architecture
		writeLine $fd $tabLvl "architecture a of $compName is"
		incr tabLvl

		# component definition - generic map, port map
		incr tabLvl
		genHdlComponent $fd $lang $compName $instName $portListSubst $genericsList $tabLvl
		incr tabLvl -1
		writeLine $fd $tabLvl "-- synthesis translate_off"
		writeLine $fd $tabLvl "-- component definition not needed by synthesis"
		incr tabLvl
		genHdlComponent $fd $lang $compName "${compName}_sim" $portListSubst [list] $tabLvl
		incr tabLvl -1
		writeLine $fd $tabLvl "-- synthesis translate_on"
		#constants
		writeLine $fd $tabLvl "-- constant used to detect synthesis vs. simulation"
		writeLine $fd $tabLvl "constant isSimulation : std_logic := '0'"
		writeLine $fd $tabLvl "-- synthesis translate_off"
		writeLine $fd $tabLvl "or '1'"
		writeLine $fd $tabLvl "-- synthesis translate_on"
		writeLine $fd $tabLvl ";"

		# signals

		incr tabLvl -1
		writeLine $fd $tabLvl "begin"
		incr tabLvl

		# assignments

		# instances - generic map, port map
		writeLine $fd $tabLvl "-- instance for synthesis. '_syn component must have been generated from MegaWizard using supplied script"
		writeLine $fd $tabLvl "alabel0: if isSimulation = '0' generate"
		incr tabLvl
		genHdlInstance $fd $lang "${compName}_syn" $instName $portListSubst $genericsList $tabLvl
		incr tabLvl -1
		writeLine $fd $tabLvl "end generate;"
		writeLine $fd $tabLvl ""
		writeLine $fd $tabLvl "-- instance for simulation, '_sim component must have been generated from MegaWizard using supplied script"

		# write out command line for generating MegaWizard component - Quartus will do this automatically, but error prone, however simulation
		writeLine $fd $tabLvl "-- Command line for generating MegaWizard component and sim model:"
		set lineStr "-- qmegawiz -silent module=\"$instName\" "
		foreach {key val} [array get mwparmsOutputParamsArr] {
			append lineStr "$key=$val "
		}
		append lineStr "-p:work_dir ${compName}_sim.vhd"
		writeLine $fd $tabLvl $lineStr

		writeLine $fd $tabLvl "-- synthesis translate_off"
		writeLine $fd $tabLvl "alabel1: if isSimulation = '1' generate"
		incr tabLvl
		genHdlInstance $fd $lang $compName "${compName}_sim" $portListSubst [list] $tabLvl
		incr tabLvl -1
		writeLine $fd $tabLvl "end generate;"
		writeLine $fd $tabLvl "-- synthesis translate_on"

		incr tabLvl -1
		writeLine $fd $tabLvl "end;"
		writeLine $fd $tabLvl ""
	}

	proc genHdlForOp {fd opName opIOType lang compName paramList} {
		# bring specific configuration variables for specified op's type into scope
		constants::import $opIOType
		namespace import constants::${opIOType}::*

		array set paramArr $paramList

		set mwParmsBaseFileName $OPNAME_TO_MWPARMS_NAME_MAP($opName)

		# open associated .mwparms files to pull in info for MegaWizard parameters replacing values accordingly
		set inMwparmsFileName [file join $MEGAWIZARD_FP_PATH "${mwParmsBaseFileName}.mwparms"]
		array set mwparmsOutputParamsArr {}
		array set replaceParmsArr [convertXmlToMwparmsParameters $paramList]
		::altera::mwparms::setParams $inMwparmsFileName "mwparms_out.unused" replaceParmsArr mwparmsOutputParamsArr

		# open associated .inst file to pull in info for module name, generics, etc. replacing values accordingly
		set inInstFileName [file join $MEGAWIZARD_FP_PATH "${mwParmsBaseFileName}.mwinst"]
		array set mwinitOutputParamsArr {}
		array set replaceParmsArr [convertXmlToMwGenericParameters $paramList]
		#puts "replaceParmsArr:"
		#parray replaceParmsArr
		::altera::mwinst::setParams $inInstFileName "mwinst_out.unused" replaceParmsArr mwinitOutputParamsArr

		set mwCompName "${compName}_${mwinitOutputParamsArr(_MODULE)}"

		set compGenerics [list]
		foreach {key val } [array get mwinitOutputParamsArr] {
			if { [string index $key 0] != "_" } {
				lappend compGenerics [list $key $val]
			}
		}

		# add ports to HDL to be generated
		hdlDefinePorts paramArr $DEF_PORTS
		hdlDefineGenerics paramArr $DEF_GENERICS
		# can't insert complicated code in component, avoid and use simple components for now and genHdlWrapperForOpComp
		#hdlDefineComponent $mwCompName paramArr $DEF_COMP_PORTS $compGenerics $opName OPNAME_TO_RESULT_PORT_MAP
		hdlDefineComponent $mwCompName paramArr $DEF_COMP_PORTS [list] $opName OPNAME_TO_RESULT_PORT_MAP
		hdlDefineSignals paramArr $DEF_SIGNALS

		# add body to HDL to be generated
		hdlInstanceComponent $mwCompName paramArr $COMP_INST_PORT_MAP $opName OPNAME_TO_RESULT_PORT_MAP

		# call additional logic generation if routine exists
		if {[info proc genIpForOpIOType] != ""} {
			genIpForOpIOType $fd $opName $opIOType $lang $compName paramArr
		}

		#	libraries are not complete and require the generated .vho file
		# write HDL for specialized wrapper
		genHdlWrapperForOpComp $fd $lang $mwCompName ${mwinitOutputParamsArr(_MODULE)} paramArr $DEF_COMP_PORTS $compGenerics $opName OPNAME_TO_RESULT_PORT_MAP mwparmsOutputParamsArr

		# write HDL for autoip function
		printComment $fd $lang "$compName is instantiated within '_comp.vhd file"
		hdl::gen $fd $compName $lang

		set ::altera::floating_point::GeneratedOpArr($mwCompName) "$paramList"

		# remove imported specific configuration (next call may need to import a different set)
		namespace forget constants::${opIOType}::*
		namespace forget constants::${opIOType}::*
	}

	proc genIp {fd lang compName paramList} {
		puts "genIp $lang $compName $paramList"

		switch -regexp $compName {
			{fadd_+} -
			{fsub_+} -
			{fmul_+} -
			{fdiv_+} -
			{faddd_+} -
			{fsubd_+} -
			{fmuld_+} -
			{fdivd_+} {
				set opName "[lindex [split $compName "_"] 0]"
				set opIOType twoFloatInOneFloatOut
			}
			{fcos_+} -
			{fsin_+} -
			{fatan_+} -
			{flog_+} -
			{flogd_+} -
			{fexp_+} -
			{fexpd_+} -
			{fsqrt_+} -
			{fsqrtd_+} {
				set opName "[lindex [split $compName "_"] 0]"
				set opIOType oneFloatInOneFloatOut
			}
			{fcmp_+} {
				set opName "f[lindex [split $compName "_"] 1]"
				set opIOType twoFloatInOneBitOut
			}
			{fcmpd_+} {
				set opName "f[lindex [split $compName "_"] 1]d"
				set opIOType twoFloatInOneBitOut
			}
			{[fd]toi_+} {
				set opName "[lindex [split $compName "_"] 0]"
				set opIOType oneFloatInOneIntOut
			}
			{ito[fd]_+} {
				set opName "[lindex [split $compName "_"] 0]"
				set opIOType oneIntInOneFloatOut
			}
			{uto[fd]_+} {
				set opName "[lindex [split $compName "_"] 0]"
				set opIOType oneUIntInOneFloatOut
			}
			default {
				fatalError "Unknown component name $compName"
			}
		}
		genHdlForOp $fd $opName $opIOType $lang $compName $paramList

		# write out current list of op's encountered
		set fout [open $::altera::floating_point::constants::FP_OP_LIST_FILENAME w]
		foreach {cmpName cmpParam} [array get ::altera::floating_point::GeneratedOpArr] {
			puts $fout "${cmpName}"
		}
		close $fout
	}

	proc calcTiming {opName field paramList} {
		#puts "calcTiming: opName=$opName field=$field paramList=$paramList"
		switch $field {
			rate {
				# return 1
				array set paramArr $paramList
				return $paramArr(rate)
			}
			cycles {
				array set paramArr $paramList
				return $paramArr(latency)
			}
		}
	}

	#####
	# initialize library on load
	init
}

return 0