###############################################################################
# Copyright (c) 2014, Impulse Accelerated Technologies, Inc.
# All Rights Reserved.
#
# Generate bus interface for Altera Qsys with FPUpdate patch installed
#
# These procedures must be called from within impulse_arch.exe.
#
###############################################################################

proc GenerateBUS {outdir name connections} {
	# load embeddedTclFile utilities
	global env
	global EmbeddedTclVars
	global dual_clock
	set strmInCount 0
	set strmOutCount 0
	set regInCount 0
	set regOutCount 0

	#TODO: Support dual clocks
	if { $dual_clock } {
		puts "genbus.tcl:GenerateBUS():ERROR: Dual clocks are currently unsupported in this PSP.  Please de-select this option in the project settings."
		exit -1
	}
	
	if { $::_flags(use_std_logic) } {
		puts "genbus.tcl:GenerateBUS():ERROR: std_logic types for VHDL interfaces are currently unsupported in this PSP.  Please de-select this option in the project settings"
		exit -1;
	}

	if { $::_flags(active_low_reset) } {
		puts "genbus.tcl:GenerateBUS():ERROR: Active low reset is currently unsupported in this PSP.  Please de-select this option in the project settings."
		exit -1;
	}
	
	
	set keyValConnectionList [psp_genbus_lib::createConKV hw $connections]
	puts "keyValConnectionList = $keyValConnectionList"
	
	# generate wrapper file from embedded Tcl wrapper file
	set EmbeddedTclVars(ArchName) "$name"
	set EmbeddedTclVars(OutputDir) "$outdir"
	set EmbeddedTclVars(KeyValConnections) $keyValConnectionList
	# original connections list necessary for using builtin functions for generating HDL
	set EmbeddedTclVars(OrigConnections) $connections

	set EmbeddedTclVars(PspVersion) "2.0.0.0"
	# copy project files and generate any from template files with embedded Tcl
	if { [catch {copyDirAndGenerateEmbeddedTclFiles "$psp_top::PspHdlBaseDir/genbus_hw" $outdir/export}] } {
		puts "genbus.tcl:Error Caught"
		global errorInfo
		puts "errorInfo=$errorInfo"
		exit -1;
	}
	
	if {$psp_top::IsGeneratingTestbench == 0} {
   	# generate any floating point operators needed
		set genDir [file join $outdir lib gen]
		puts "\nINFO:Generating IP cores using MegaWizard from command line, this may take a few minutes..."
		
		set subInterp [interp create]
		interp alias $subInterp exit {} interp delete $subInterp  
		#subInterpretExit $subInterp

		#		if { [interp eval $subInterp "cd $genDir;source {gen_vhdl.tcl}"] != 0 } 
		if { [catch {$subInterp eval "cd $genDir;source {gen_vhdl.tcl}"} ] } {
			puts "\nError encountered while generating IP cores using MegaWizard from command line."
			puts "You may need to manually run the gen_vhdl.bat/.sh script located in $genDir before export and/or simulation" 
		} else {
			puts "\nINFO:Done generating IP cores using MegaWizard from command line."
		}
	}
	
	# Software files generated during genbus stage are handled separately so GenerateLib can
	#  copy them over to the SW gen directory so SW export can finally copy them to SW target dir
#	if { [catch {copyDirAndGenerateEmbeddedTclFiles "$psp_top::PspHdlBaseDir/genbus_sw" $outdir/genbus_sw}] } {
#		puts "genbus.tcl:Error Caught"
#		global errorInfo
#		puts "errorInfo=$errorInfo"
#		exit -1;
#	}

}
